/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *	Mark Alexander
 *	Side Effects Software Inc
 *	477 Richmond Street West
 *	Toronto, Ontario
 *	Canada   M5V 3E7
 *	416-504-9876
 *
 * NAME:	CHOP_Puppet.C ( CHOP Library, C++)
 *
 * COMMENTS:
 */


#include <strstream.h>
#include <iostream.h>
#include <fstream.h>
#include <unistd.h>
#include <fcntl.h>

#include <UT/UT_Serial.h>
#include <UT/UT_SysClone.h>
#include <UT/UT_Defines.h>
#include <UT/UT_CPIO.h>
#include <UT/UT_Math.h>
#include <UT/UT_Intercept.h>
#include <UT/UT_Interrupt.h>
#include <UT/UT_Version.h>

#include <CH/CH_LocalVariable.h>
#include <OP/OP_Operator.h>
#include <OP/OP_Director.h>
#include <OP/OP_Network.h>
#include <PRM/PRM_Include.h>

#include <CHOP/CHOP_Node.h>
#include <CHOP/PRM_ChopShared.h>

#include "CHOP_Puppet.h"

//////////////////////////////////////////////////////////////////////////

#include <UT/UT_DSOVersion.h>
#include <OP/OP_OperatorTable.h>

Export void
newChopOperator(OP_OperatorTable *table)
{
    table->addOperator(
	    new OP_Operator("puppet", "Puppet",
		 CHOP_Puppet::myConstructor,
		&CHOP_Puppet::myTemplatePair, 0, 0,
		&CHOP_Puppet::myVariablePair,
		OP_FLAG_GENERATOR)
    );
}

//////////////////////////////////////////////////////////////////////////

#define PUPPET_NOT_CONNECTED	0
#define PUPPET_OKAY		1
#define PUPPET_UPSET		2

#define REPORT_SINGLE		0
#define REPORT_CONTINUOUS	1

#define TWO_BYTE_ROT		0
#define THREE_BYTE_ROT		1

// We want to save in ~/houdini6.x/*.pref, where 6.x is the current
// version number.
#define PUPPET_PREF_FILE	"$HOME/houdini" UT_SAVE_VERSION "/puppetworks.pref"
#define PUPPET_PREF_HEADER	"Puppet"

OP_TemplatePair CHOP_Puppet::myTemplatePair(
    CHOP_Puppet::myTemplateList, &CHOP_Node::myTemplatePair);

OP_VariablePair CHOP_Puppet::myVariablePair(
    CHOP_Puppet::myVariableList, &CHOP_Node::myVariablePair);

CHOP_SWITCHER4(4,"Puppet", 3, "Calibrate", 4,"Port", 3,"Channel");

static PRM_Name        names[] =
{
    PRM_Name("active",		"Active"),
    PRM_Name("mode",		"Record Mode"),
    PRM_Name("rate",		"Report Rate"),
    PRM_Name("usecycle",	"Use Cycles"),
    PRM_Name("grabzero",	"Grab Zero Position"),
    PRM_Name("usezero",		"Use Zero Position"),
    PRM_Name("writezero",	"Write Zero Position To Preferences"),

    PRM_Name("port",		"Port"),
    PRM_Name("portname",	"Port Name"),
    PRM_Name("baudrate",	"Baud Rate"),
    PRM_Name("reset",		"Reset Puppet"),
};


static PRM_Name modeItems[] = {
    PRM_Name("current",		"Current Frame"),
    PRM_Name("timeslice",	"Current Time Slice"),
    PRM_Name(0),
};

static PRM_ChoiceList modeMenu((PRM_ChoiceListType)
    (PRM_CHOICELIST_EXCLUSIVE | PRM_CHOICELIST_REPLACE), modeItems);

static PRM_Name reportItems[] = {
    PRM_Name("r240",	"240 Hz"),
    PRM_Name("r180",	"180 Hz"),
    PRM_Name("r120",	"120 Hz"),
    PRM_Name("r100",	"100 Hz"),
    PRM_Name("r60",	"60 Hz"),
    PRM_Name("r50",	"50 Hz"),
    PRM_Name("r30",	"30 Hz"),
    PRM_Name("r25",	"25 Hz"),
    PRM_Name("r24",	"24 Hz"),
    PRM_Name("r20",	"20 Hz"),
    PRM_Name("r15",	"15 Hz"),
    PRM_Name("r10",	"10 Hz"),
    PRM_Name("r5",	"5 Hz"),
    PRM_Name("r3",	"3 Hz"),
    PRM_Name("r2",	"2 Hz"),
    PRM_Name("r1",	"1 Hz"),
    PRM_Name(0),
};

static PRM_ChoiceList reportMenu((PRM_ChoiceListType)
    (PRM_CHOICELIST_EXCLUSIVE | PRM_CHOICELIST_REPLACE), reportItems);

static int reportRate [] = { 240,180,120,100,60,50,30,25,24,20,15,10,5,3,2,1 };
static PRM_Default reportDef(6);

static PRM_Name portItems[] = {
    PRM_Name("serial1",	"Serial Port 1"),
    PRM_Name("serial2",	"Serial Port 2"),
    PRM_Name("serial3",	"Serial Port 3"),
    PRM_Name("serial4",	"Serial Port 4"),
    PRM_Name("other",	"Other Port"),
    PRM_Name(0),
};
#define OTHER_PORT 4

static PRM_ChoiceList portMenu((PRM_ChoiceListType)
    (PRM_CHOICELIST_EXCLUSIVE | PRM_CHOICELIST_REPLACE), portItems);

static const char * serialPort[] =
{
    "/dev/ttyd1",
    "/dev/ttyd2",
    "/dev/ttyd3",
    "/dev/ttyd4",
};

static PRM_Default portNameDef(0,"/dev/ttyd5");

static PRM_Name baudItems[] = {
    PRM_Name("b9600",	"9.6K"),
    PRM_Name("b19200",	"19.2K"),
    PRM_Name("b38400",	"38.4K"),
    PRM_Name(0),
    PRM_Name("b57600",	"57.6K"),  // not supported under 6.2.. maybe later
    PRM_Name("b115k",	"115.2K"),
};

#define NUM_BAUD_RATES 3
static int baudrate[] = { 9600, 19200, 38400, 57600, 115200 };

static PRM_ChoiceList baudMenu((PRM_ChoiceListType)
    (PRM_CHOICELIST_EXCLUSIVE | PRM_CHOICELIST_REPLACE), baudItems);

PRM_Template
CHOP_Puppet::myTemplateList[] =
{
    PRM_Template(PRM_SWITCHER,  5, &PRMswitcherName, switcher),

    // puppet
    PRM_Template(PRM_TOGGLE,	1, &names[0], PRMoneDefaults),
    PRM_Template(PRM_ORD,	1, &names[1], PRMzeroDefaults, &modeMenu),
    PRM_Template(PRM_ORD,	1, &names[2], &reportDef, &reportMenu),
    PRM_Template(PRM_TOGGLE,	1, &names[3], PRMzeroDefaults),

    // calibrate
    PRM_Template(PRM_CALLBACK,	1, &names[4], 0,0,0, &CHOP_Puppet::grabZero),
    PRM_Template(PRM_TOGGLE,	1, &names[5], PRMoneDefaults),
    PRM_Template(PRM_CALLBACK,  1, &names[6], 0,0,0, &CHOP_Puppet::writeZero),
    
    // port
    PRM_Template(PRM_ORD,	1, &names[7], PRMzeroDefaults, &portMenu, 0,
		 portChanged),
    PRM_Template(PRM_STRING,    1, &names[8], &portNameDef, 0, 0,
		 portChanged),
    PRM_Template(PRM_ORD,	1, &names[9], PRMtwoDefaults, &baudMenu,0,
		 portChanged),
    PRM_Template(PRM_CALLBACK,  1, &names[10], 0, 0, 0, CHOP_Puppet::reset),

    // channel
    PRM_Template(PRM_ORD,	1, &CHOP_ExtendLeftName,0, &CHOP_ExtendMenu),
    PRM_Template(PRM_ORD,	1, &CHOP_ExtendRightName, 0, &CHOP_ExtendMenu),
    PRM_Template(PRM_FLT_J,	1, &CHOP_DefaultValueName, 0, 0,
		 &CHOP_DefaultValueRange),

    PRM_Template()
};

CH_LocalVariable
CHOP_Puppet::myVariableList[] = {
    { 0, 0, 0 }
};

OP_Node *
CHOP_Puppet::myConstructor(OP_Network	*net,
			 const char	*name,
			 OP_Operator	*op)
{
    return new CHOP_Puppet(net, name, op);
}

//----------------------------------------------------------------------------

#define HEADER1 0xC1 
#define HEADER2 0xC2
#define HEADER3 0xC3
#define TRAILER 0xC0




CHOP_Puppet::CHOP_Puppet( OP_Network	*net,
		      const char	*name,
		      OP_Operator	*op)
	   : CHOP_RealtimeGenerator(net, name, op)
{
    myParmBase    = getParmList()->getParmIndex( names[0].getToken() );
    
    myBaudRate    = -1;
    myPuppet	  = 0;
    myReportRate  = 0;
    myReportMode  = 0;
    myNumJoints   = 0;
    myActiveState = 0;
    myDemandSent  = 0;
    myVersionMajor= 0;
    myVersionMinor= 0;
    myTimeStamp	  = -1;
    myExpectedStamp = -1;
    myRecordData  = 0;
    myAngleFormat = -1;
    mySerialPos	  = 0;
    mySerialSize  = 0;
    mySerialOffset = 0;
    myPuppetState = PUPPET_NOT_CONNECTED;
    myConfigNumJoints = 0;
    myConfigInfo      = 0;
    myIsConfigured    = 0;
    myUseConfig	      = 0;
    myBoss	      = 0;

    loadPreferenceConfig();
}

CHOP_Puppet::~CHOP_Puppet()
{
    closePort();
}

unsigned
CHOP_Puppet::disableParms()
{
    unsigned changes = CHOP_Node::disableParms();

    if(LEXTEND() == CL_TRACK_DEFAULT ||
       REXTEND() == CL_TRACK_DEFAULT)
	changes += enableParm(CHOP_DefaultValueName.getToken(), 1);
    else
	changes += enableParm(CHOP_DefaultValueName.getToken(), 0);

    changes += enableParm("portname", PORT() == OTHER_PORT);
    return changes;
}

void
CHOP_Puppet::realtimeStart(float )
{
    if(!UTgetIntercept()->getPollMode())
    	myReportMode = REPORT_SINGLE;
    else
	myReportMode = REPORT_CONTINUOUS;

    flushPort();
    if(myReportMode)
    {
	sendTimedDataCommand();
	myTimeStamp = -1;
	myExpectedStamp = -1;
    }
    else
    {
	sendSingleDataCommand();
	sginap(1);
	sendGetDataCommand();
    }
}

void
CHOP_Puppet::realtimeStop()
{
    /// stop timer mode
    if(myReportMode)
	sendSingleDataCommand();

    myDemandSent = 0;
    
    // flush the input buffer.
    flushPort();
}    


OP_ERROR
CHOP_Puppet::cookMyChop(OP_Context &)
{
    // this is never called. ever. (really.) 
    return error();
}

OP_ERROR
CHOP_Puppet::cookMySlice(OP_Context &context, int start, int end)
{
    int active,report,usecycles;

    updateConnection();
    
    if(myPuppetState == PUPPET_UPSET)
    {
	// try to reinit
	sendInitialize();
	realtimeStart(context.myTime);
    }
    else
    {
	active = ACTIVE();
	
	// if 'use cycles' is switched
	usecycles = USE_CYCLES();
	if(usecycles != myAngleFormat)
	{
	    sendSingleDataCommand();
	    flushPort();
	    sendAngleFormatCommand(usecycles);

	    if(active)
		realtimeStart(context.myTime);
	}
	
	// turn off if active turned off.
	if(!active && myActiveState)
	    realtimeStop();
	else if(active && !myActiveState)
	    realtimeStart(context.myTime);

	myActiveState = active;
	
	// start/stop framebar stuff: single mode while scrubbing
	if(UTgetIntercept()->getPollMode() && myReportMode == REPORT_SINGLE)
	    realtimeStart(context.myTime);
	else if(!UTgetIntercept()->getPollMode() &&
		myReportMode == REPORT_CONTINUOUS)
	{
	    realtimeStop();
	    realtimeStart(context.myTime);
	}
	     

	// if 'report rate' is switched 
	report = REPORT_RATE();
	if(myReportRate != report)
	{
	    myReportRate = report;
	    sendReportRateCommand();
	}


	myUseConfig = (USE_ZERO() & myIsConfigured);
    }
    
    if(myPuppetState == PUPPET_OKAY)
    {
	if(myTimeStamp != -1)
	{
	    if(MODE())
		myExpectedStamp = myTimeStamp+1;
	    else
		myExpectedStamp = myTimeStamp + (end-start);
	}
	else
	    myExpectedStamp = -1;
		
	if(myReportMode == REPORT_SINGLE)
	    recordSingle();
	else
	    recordTimeSlice();
    }
    else 
	addError(CHOP_ERROR_DEVICE_NOT_FOUND);

    UTaddMessage("CHOP", CHOP_MESSAGE_PUPPET_VERSION, myVersionString);
    return error();
}

void
CHOP_Puppet::recordSingle()
{
    int		 i,j;
    ut_PuppetData *block;
    int		 len = myClip->getTrackLength();
    CL_Track	*track;
    float	*towrite;
    float	 weight;


    if(getAngles(myNumJoints*2, myRecordData, myTimeStamp, 1))
    {
//	cerr << "Time Stamp at " << myTimeStamp << endl;
	for(i=0; i<myNumJoints*2; i++)
	{
	    block = (ut_PuppetData *) getDataBlock(i);
	    track = myClip->getTrack(i);
	    towrite = track->getData();

	    if(len > 1)
	    {
		for(j=0; j<len; j++)
		{
		    weight = float(j+1) / len;
		    
		    towrite[j] = weight * myRecordData[i] +
			(1.0F-weight)*block->myLastValue;
		}
	    }
	    else
		towrite[0] = myRecordData[i];
	    
	    block->myLastValue = myRecordData[i];
	}
    }
    else
    {
	for(i=0; i<myNumJoints*2; i++)
	{
	    block = (ut_PuppetData *) getDataBlock(i);
	    track = myClip->getTrack(i);
	    towrite = track->getData();
	    
	    for(j=0; j<len; j++)
		towrite[j] = block->myLastValue;
	}
    }
	    
    sendGetDataCommand();
    
}

void
CHOP_Puppet::recordTimeSlice()
{
    if(!myRecordData || !myPuppet)
	return ;
    
    int		 length = myClip->getTrackLength();
    int		 ts;
    int		 i,j;
    int		 end;
    float	 *data;
    float	 value;
    int		 first = 1;
    int		 datanum = length;
    
    for(i=0; i<length; i++)
    {
	if(!getAngles(myNumJoints*2,myRecordData,ts,datanum,first))
	    break;
	

	for(j=myNumJoints*2; j--;)
	{
	    ((ut_PuppetData *)getDataBlock(j))->myLastValue =
		myRecordData[j];
	    myClip->getTrack(j)->getData()[i] = myRecordData[j];
	}
	
// 	myTimeStamp = ts;
// 	myExpectedStamp = myTimeStamp + 1;
//	first = 0;
	datanum--;
    }

    end = i;

    if(end < length)
    {
	for(j=myNumJoints*2; j--;)
	{
	    value = ((ut_PuppetData *)getDataBlock(j))->myLastValue;
	    data = myClip->getTrack(j)->getData();
	
	    for(i=end; i<length; i++)
		data[i] = value;
	}
    }
    
    myTimeStamp = ts;
}

 


void
CHOP_Puppet::updateJointTracks()
{
    UT_PtrArray<char *> names;
    
    getChannelOutputNames(names);
    cacheChannelData(0);
    
    if(myRecordData)
	delete [] myRecordData;
    
    myRecordData = new float[ myNumJoints*2 ];
}

void
CHOP_Puppet::getChannelOutputNames(UT_PtrArray<char *> &str)
{
    UT_String s;
    char *name;
    char  number[32];
    int i;

    str.entries(myNumJoints*2);
    str.resize(myNumJoints*2);
    
    for(i=0; i<myNumJoints; i++)
    {
	UT_String::itoa(number,i+1);

	s = number;
	s+= "a";
	name = new char[s.length()+1];
	strcpy(name, (const char *)s);

	str(i*2) = name;

	s = number;
	s+= "b";
	name = new char[s.length()+1];
	strcpy(name, (const char *)s);

	str(i*2+1) = name;
    }
}

int
CHOP_Puppet::realtimeMode()
{
    return MODE() ? REALTIME_MODE_TRIM : REALTIME_MODE_CURRENT_FRAME;
}

float
CHOP_Puppet::getRealtimeSampleRate()
{
    return reportRate[ REPORT_RATE() ];
}

int
CHOP_Puppet::grabZero(void *data, int, float, const PRM_Template * )
{
    CHOP_Puppet *me = (CHOP_Puppet *) data;

    me->grabZeroPosition();
    return 1;
}

int
CHOP_Puppet::writeZero(void *data, int, float, const PRM_Template * )
{
    CHOP_Puppet *me = (CHOP_Puppet *) data;

    me->savePreferenceConfig();
    return 1;
}

int
CHOP_Puppet::reset(void *data, int, float, const PRM_Template * )
{
    CHOP_Puppet *me = (CHOP_Puppet *) data;

    me->sendFrameResetCommand();
    me->myPort.harden("");
    me->updateConnection();

    return 1;
}

int
CHOP_Puppet::portChanged(void *data, int, float, const PRM_Template *)
{
    CHOP_Puppet *me = (CHOP_Puppet *) data;

    me->updateConnection();

    return 1; 
}

ut_RealtimeData *
CHOP_Puppet::newRealtimeDataBlock(const char *name, const CL_Track *,
				  float )
{
    return new ut_PuppetData(name);
}

void
CHOP_Puppet::updateConnection()
{
    UT_String	portname;
    int		baudr;
    int		port;

    port = PORT();
    if(port == OTHER_PORT)
	PORT_NAME(portname);
    else
	portname.harden(serialPort[port]);

    baudr = BAUD_RATE();
    
    if(myPort != portname ||
       myPuppetState != PUPPET_OKAY)
    {
	myPort.harden(serialPort[port]);
	openPort();
    }
    else if(baudr != myBaudRate && myPuppetState == PUPPET_OKAY)
    {
	sendSingleDataCommand();
	flushPort();
	changeBaud(baudr);
	realtimeStart(0);
    }
}
    
void
CHOP_Puppet::openPort()
{
    int i;
    closePort();

//    cerr << "Opening port" << endl;

    myPuppetState = PUPPET_NOT_CONNECTED;
    myPuppet = new UT_Serial(baudrate[0], 8);
    
    if(myPuppet->open(myPort, O_RDWR))
    {
	for(i=0; i<NUM_BAUD_RATES; i++)
	{
	    flushPort();
	    sendSingleDataCommand();
	    
//	    cerr << "Trying " << baudrate[i] << " BAUD" << endl;
	    myBaudRate = i;
	    if(getVersion())
	    {
		myPuppetState = PUPPET_OKAY;
		sendInitialize();
		
		if(myPuppetState == PUPPET_OKAY)
		    break;
	    }
		    
	    if(i+1 < NUM_BAUD_RATES)
		myPuppet->changeRate(baudrate[i+1]);
	}
    }

    if(myPuppetState != PUPPET_OKAY)
	closePort();
}

void
CHOP_Puppet::flushPort()
{
    if(myPuppet)
	myPuppet->flush();
	
    mySerialOffset = 0;
    mySerialPos = 0;
    mySerialSize = 0;

}

void
CHOP_Puppet::closePort()
{
    flushPort();
    
    if(myPuppet)
    {
	sendSingleDataCommand();
	changeBaud(0);
	delete myPuppet;
    }
    
    myPuppet = 0;
    myNumJoints = 0;
    myPuppetState = PUPPET_NOT_CONNECTED;
    myDemandSent = 0;
}

void
CHOP_Puppet::changeBaud(int b)
{
    char baud[2];
    int	 count, ack;

    if(myBaudRate != b)
    {
	// set BAUD Rate
	baud[0] = 'B';
	baud[1] = b;

//	cerr << "Changing the baud rate" << endl;
	sendPacket(HEADER1, baud, 2);


	count = 20;
	ack = 0;
	
	while(!(ack = getAcknowledge()) && count--)
	{
	    sginap(1);
	    if(myBoss && myBoss->opInterrupt())
		break;
	}

	if(!ack)
	{
//	    cerr << "Couldn't change rate!!" << endl;
	    return;
	}

	if(!myPuppet->changeRate(baudrate[b]))
	{
//	    cerr << "BAUD RATE problems " << endl;
	    myPuppetState = PUPPET_NOT_CONNECTED;
	}
//	sginap(10);

	myPuppet->flush();
    }
    myBaudRate = b;
}
    

void
CHOP_Puppet::sendPacket(char header, char *msg, int num)
{
    if(!myPuppet)
	return;
    
    int chksum = header;
    char cs,trail = TRAILER;
    int	 i;

    for(i=0; i<num; i++)
	chksum += msg[i];

    cs = (char) (chksum&0x0000007f);
    myPuppet->write(&header, 1);
    myPuppet->write(msg, num);
    myPuppet->write(&cs, 1);
    myPuppet->write(&trail, 1);
    
/*  cerr << "Sending " << (int) header << "  ";

    for(i=0; i<num; i++)
	cerr << (int) msg[i] << " ";

    cerr << (int)cs << " " << (int)trail << endl;
    */
}

int
CHOP_Puppet::receivePacket(int wait, int recentangles)
{
    if(!myPuppet)
	return 0;
    
    int		read;
    int		chksum;
    int		i,pos,psize;
    int		start,end;
    int		found = 0;
    int		toread;
    int		count = wait;
    int		type;
    int		recentcount,laststart,lastend;

//    cerr << "RP " << endl;
    do
    {
	// read up to SERIAL_BUF_SIZE bytes
	toread = SERIAL_BUF_SIZE - mySerialSize;
	
	if(mySerialOffset + toread >= SERIAL_BUF_SIZE)
	{
	    read = myPuppet->read(mySerialBuffer+mySerialOffset,
				  SERIAL_BUF_SIZE-mySerialOffset);

	    if(read == SERIAL_BUF_SIZE-mySerialOffset && read < toread)
		read += myPuppet->read(mySerialBuffer, toread-read);
	}
	else
	{
	    read = myPuppet->read(mySerialBuffer+mySerialOffset, toread);
	}
//  	cerr << "Read " << read << " " << toread << endl;
//  	cerr << "Size, Pos, Offset = " << mySerialSize << " " << mySerialPos << " " << mySerialOffset << endl;
	mySerialOffset += read;
	if(mySerialOffset >=SERIAL_BUF_SIZE)
	    mySerialOffset -= SERIAL_BUF_SIZE;
    
	mySerialSize += read;

	found = 0;
	if(recentangles)
	{
	    recentcount = recentangles;
	    for(i=0, pos = mySerialOffset; i<mySerialSize; i++)
	    {
		if(pos == 0)
		    pos = SERIAL_BUF_SIZE-1;
		else
		    pos--;

		if(mySerialBuffer[pos] == TRAILER)
		{
		    found = 1;
		    end = pos;
		}
		else if(mySerialBuffer[pos]>TRAILER &&
			mySerialBuffer[pos]<=HEADER3 && found)
		{
		    if(type == 'j' || type=='J')
		    {
			found = 2;
			start = pos;
			recentcount --;
			
			if(recentcount == 0 )
			{
//			    cerr << "correct ";
			    break;
			}
			else
			{
			    laststart = start;
			    lastend = end;
			}
			     
		    }
		    found = 0;
		}
		type = mySerialBuffer[pos];
	    }

	    if(recentcount < recentangles && found != 2)
	    {
//		cerr << "later " << recentcount << " " ;
		found = 2;
		start = laststart;
		end = lastend;
	    }
	}
	else
	{
	    for(i=0, pos = mySerialPos; i<mySerialSize; i++)
	    {
		if(mySerialBuffer[pos] > TRAILER && mySerialBuffer[pos] <= HEADER3)
		{
		    found = 1;
		    start = pos;
		}
		else if(mySerialBuffer[pos] == TRAILER && found)
		{
		    found = 2;
		    end = pos;
		    break;
		}

		pos++;
		if(pos == SERIAL_BUF_SIZE)
		    pos = 0;
	    }
	}

	if(found == 2)
	{
	    if(end > start)
	    {
		memcpy(myBuffer, mySerialBuffer+start, end-start+1);
		psize = end-start+1;
	    }
	    else
	    {
		memcpy(myBuffer, mySerialBuffer+start, SERIAL_BUF_SIZE-start);
		memcpy(myBuffer+(SERIAL_BUF_SIZE-start),mySerialBuffer, end+1);

		psize = (SERIAL_BUF_SIZE-start + end+1);
	    }

	    mySerialPos = end+1;
	    if(mySerialPos >= SERIAL_BUF_SIZE)
		mySerialPos -= SERIAL_BUF_SIZE;
	    
	    if(mySerialOffset >= mySerialPos)
		mySerialSize = mySerialOffset-mySerialPos;
	    else
		mySerialSize = (mySerialOffset)+(SERIAL_BUF_SIZE-mySerialPos); 
	

	    chksum = 0;
	    for(i=0; i<psize-2; i++)
		chksum += myBuffer[i];

	    if((chksum&0x0000007F) == (int)myBuffer[psize-2])
		return 1;
// 	    else
// 		cerr << "Bad CHECKSUM " << endl;
	}

	if(!wait)
	    break;

	sginap(1);
//	cerr << "Nap" << endl;
	
    } while(count -- > 0);
	
    return 0;
}

int
CHOP_Puppet::getVersion()
{
    char command = 'V';
    int	 count;
    int  rcvd = 0;
    char version[256];

    sendPacket(HEADER1, &command, 1);

    while(!rcvd)
    {
	count = 2;
	while(!(rcvd = receivePacket()) && count--)
	    sginap(1);
	
	if(rcvd)
	{
	    if(myBuffer[1] == 'V')
	    {
		myVersionMajor = myBuffer[3];
		myVersionMinor = myBuffer[2];

		sprintf(version, " %d.%02d",
			myVersionMajor, myVersionMinor);
		myVersionString.harden(version);
	    }
	    else
		rcvd = 0;
	}
	else
	    break;
    }

    return rcvd;
}
    

int
CHOP_Puppet::getAcknowledge(int /*acknum*/)
{
    int count = 40;
    while(count --)
    {
	if(receivePacket() && myBuffer[1] == 'A')// && myBuffer[2] == acknum)
	{
//	    cerr << "Acknowledge with " << (int)myBuffer[2] << endl;
	    return 1;
	}
// 	else
// 	    cerr << "packet =  " << (int)myBuffer[0] << " " << (int)myBuffer[1]
// 		 << (int)myBuffer[2] << " " << (int)myBuffer[3] << endl;
    }
    
    return 0;
}

void
CHOP_Puppet::sendInitialize()
{
    myBoss = UTgetInterrupt();
    myBoss->opStart("Initializing PuppetWorks Device");
    
    flushPort();

    sendSingleDataCommand();
    sendNoTimeStampCommand();
    sendAngleFormatCommand(USE_CYCLES());
    changeBaud(BAUD_RATE());

    if(!sendFindJointsCmd())
	myPuppetState = PUPPET_UPSET;
    else
	myPuppetState = PUPPET_OKAY;

    myBoss->opEnd();
    myBoss = 0;
}

int
CHOP_Puppet::sendFindJointsCmd()
{
    char findjoint = 'W';
    int  count,rcvd;

//    cerr << "Sending joint request" << endl;
    sendPacket(HEADER3, &findjoint, 1);

    count = 20;

    do
    {
	rcvd = receivePacket();
	if(rcvd && myBuffer[1] == 'w')
	    break;
	
	count --;
	sginap(40);
	
	if(myBoss && myBoss->opInterrupt())
	    break;
	 
    }
    while(count);
    

    if(rcvd && myBuffer[1] == 'w')
    {
	myNumJoints = myBuffer[2];
	updateJointTracks();
	return 1;
    }
    
//    cerr << "No joints " << endl;
    return 0;
}

void
CHOP_Puppet::sendReportRateCommand()
{
    char buffer[2];
    
//    cerr << "Sending report rate CMD " << endl;

    // set timer rate
    buffer[0] = 'T';
    buffer[1] = REPORT_RATE();

    sendPacket(HEADER2, buffer, 2);

    sginap(2);
    getAcknowledge();
    myReportRate = REPORT_RATE();
 }

void
CHOP_Puppet::sendTimedDataCommand()
{
    char buffer[2];

//    cerr << "Sending timed data CMD " << endl;

    // send timer mode 
    buffer[0] = 'm';
    buffer[1] = 2;
    sendPacket(HEADER2, buffer, 2);
    
    sginap(2);
    if(getAcknowledge() != 1)
	myPuppetState = PUPPET_UPSET;
    myDemandSent = 0;
}

void
CHOP_Puppet::sendSingleDataCommand()
{
    char buffer[2];

    buffer[0] = 'm';
    buffer[1] = 0;

//    cerr << "Send Single Data Mode   ";
    sendPacket(HEADER2, buffer, 2);
    
    sginap(2);
    
    if(getAcknowledge() != 1)
    {
	myPuppetState = PUPPET_UPSET;
//	cerr << "Not Set " << endl;
    }
}

void
CHOP_Puppet::sendGetDataCommand()
{
    char buffer[2];

//    cerr << "Send get data ";
    buffer[0] = 'x';
    buffer[1] = 0;

    sendPacket(HEADER2, buffer, 2);
    myDemandSent = 1;
}

void
CHOP_Puppet::sendNoTimeStampCommand()
{
    char buffer[2];

//    cerr << "Send get data ";
    buffer[0] = 'S';
    buffer[1] = 0;

    sendPacket(HEADER1, buffer, 2);

    sginap(2);
    getAcknowledge();
}


void
CHOP_Puppet::sendFrameResetCommand()
{
    char buffer[2];

//    cerr << "Send get data ";
    buffer[0] = 'R';

    sendPacket(HEADER3, buffer, 1);

    sginap(50);
    getAcknowledge();
}

void
CHOP_Puppet::sendAngleFormatCommand(int bytes)
{
    char buffer[2];

    if(myAngleFormat != bytes)
    {
	myAngleFormat = bytes;
    
	buffer[0] = 'h';
	buffer[1] = myAngleFormat ? 1 :0;

	sendPacket(HEADER3, buffer, 2);

	sginap(10);
	if(!getAcknowledge())
	{
	    cerr << "NOT SET" << endl;
	}
    }
}   


int
CHOP_Puppet::getAngles(int num,float *abuf, int & /*ts*/,int mostrecent,
		       int first)
{
    if(!myPuppet)
	return 0;
    
    int  n,i,off,index,j;
    float rot;
    
    off = 0;
    
    if(receivePacket(first?4:0, mostrecent))
    {
	if(myBuffer[1] == 'J')
	{
	    index = myBuffer[2];
	    n = myBuffer[3];

	    if(myAngleFormat == THREE_BYTE_ROT)
	    {
		for(i=0,j=4; index<num && i<n; i++,index++,j+=3)
		{
		    rot=myBuffer[j]*6.25e-5 +myBuffer[j+1]*8.0e-3 +myBuffer[j+2];
		    rot *=360.0F;
		    if(rot > 64*360)
			rot -= 128*360;

		    if(myUseConfig)
			rot -= myConfigInfo[index];
		
		    abuf[index] = rot;
		}
	    }
	    else
	    {
		for(i=0,j=4; index<num && i<n; i++,index++,j+=2)
		{
		    rot=myBuffer[j]*6.25e-5 +myBuffer[j+1]*8.0e-3;

		    rot*=360.0F;
		    if(rot > 180.0F)
			rot -= 360.0F;

		    if(myUseConfig)
			rot -= myConfigInfo[index];
		
		    if(fabs(rot) > 180.0F)
			rot = fmodf(rot+180.0F,360.0F) - 180.0F;
		    
		    abuf[index] = rot;
		}
	    }
	    off = 1;
	}
	else if(myBuffer[1] == 'j')
	{
	    index = myBuffer[2];
	    n = 1;

	    if(myAngleFormat == THREE_BYTE_ROT)
	    {
		rot = myBuffer[3]*6.25e-5 + myBuffer[4]*8.0e-3 + myBuffer[5];
		
		rot *=360.0F;
		if(rot > 64*360)
		    rot -= 128*360;
	    
		if(myUseConfig)
		    rot -= myConfigInfo[index];
	    
		abuf[index] = rot;
	    }
	    else
	    {
		rot = myBuffer[3]*6.25e-5 + myBuffer[4]*8.0e-3;
		
		rot *=360.0F;
		if(rot > 180.0F)
		    rot -= 360.0F;
	    
		if(myUseConfig)
		    rot -= myConfigInfo[index];
	    
		if(fabs(rot) > 180.0F)
		    rot = fmodf(rot+180.0F,360.0F) - 180.0F;
			
		abuf[index] = rot;
	    }
	    off = 1;
	}
    }
        
    myDemandSent = 0;

    return off;
}

void
CHOP_Puppet::grabZeroPosition()
{
    int			 i;
    float		*last;
    ut_PuppetData	*block;
    
    if(myPuppetState == PUPPET_OKAY && myNumJoints > 0)
    {
	if(myConfigNumJoints != myNumJoints || !myConfigInfo)
	{
	    if(myConfigInfo)
		delete [] myConfigInfo;
	    myConfigInfo = new float[myNumJoints*2];
	    last = 0;
	}
	else
	    last = myConfigInfo;
	
	myConfigNumJoints = myNumJoints;

	for(i=0; i<myClip->getNumTracks(); i++)
	{
	    block = (ut_PuppetData *) getDataBlock(i);

	    myConfigInfo[i] = block->myLastValue + (last ? last[i] : 0.0F);
	    block->myLastValue = 0.0F;
	}
	myIsConfigured = 1;
    }

}

void
CHOP_Puppet::savePreferenceConfig()
{
    int		i;
    int		version = 1;
    OP_Director *opdir = OPgetDirector();
    UT_String	pathname(PUPPET_PREF_FILE);

    if(opdir)
	opdir->getCommandManager()->expandString(PUPPET_PREF_FILE,
						 pathname, 0);
    
    ofstream	os((const char *)pathname);

    if(os.rdbuf()->is_open())
    {
	os << PUPPET_PREF_HEADER << endl;
	os << "Version \t" << version << endl;
	
	os << "Configured\t" << myIsConfigured << endl;
	if(myIsConfigured)
	{
	    os << "Joints    \t" << myConfigNumJoints << endl;
	    for(i=0; i<myConfigNumJoints; i++)
	    {
		os << myConfigInfo[i*2] << "\t" << myConfigInfo[i*2+1] << endl;
	    }
	}

	os.close();
    }
}



void
CHOP_Puppet::loadPreferenceConfig()
{
    char	header[1024];
    int		version;
    int		i;
    OP_Director *opdir = OPgetDirector();
    UT_String	pathname(PUPPET_PREF_FILE);

    if(opdir)
	opdir->getCommandManager()->expandString(PUPPET_PREF_FILE,
						 pathname, 0);

    ifstream is((const char *)pathname);
    
    if(is.rdbuf()->is_open())
    {
	is >> header;
	if(!strcmp(header,PUPPET_PREF_HEADER))
	{
	    is >> header;   // version
	    is >> version;

	    is >> header;  // configured
	    is >> myIsConfigured;
	    if(myIsConfigured)
	    {
		is >> header; // joints
		is >> myConfigNumJoints;

		if(myConfigNumJoints)
		{
		    if(myConfigInfo)
			delete [] myConfigInfo;
		    
		    myConfigInfo = new float[myConfigNumJoints*2];
		}
		
		for(i=0; i<myConfigNumJoints; i++)
		{
		    is >> myConfigInfo[i*2];
		    is >> myConfigInfo[i*2+1];
		}
	    }
	}
	
	is.close();
    }
}
