/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *      Side Effects
 *      477 Richmond Street West
 *      Toronto, Ontario
 *      Canada   M5V 3E7
 *      416-504-9876
 *
 * NAME:        IFD library (C++)
 *
 * COMMENTS:    Output an IFD to custom renderer
 */

#include <sys/time.h>
#include <UT/UT_DSOVersion.h>
#include <UT/UT_Matrix4.h>
#include <GEO/GEO_IOBpoly.h>
#include <GU/GU_Detail.h>
#include <CH/CH_Manager.h>
#include <OP/OP_Director.h>
#include <IFD/IFD_Scene.h>
#include <IFD/IFD_Entry.h>
#include "IFD_CustomRender.h"

//
// The renderName here should not contain any spaces.  It is used as
//	an identifier.
static const char	*renderName = "CustomRender";

//
// The constructor and destructor are relativly simple.
IFD_CustomRender::IFD_CustomRender() {}
IFD_CustomRender::~IFD_CustomRender() {}

const char *
IFD_CustomRender::renderer(void) const { return renderName; }

void
IFD_CustomRender::getOutputFlags(IFD_OutputFlags &flags) const
{
    flags.renderImage		= 1;	// Generates an image
    flags.renderScript		= 1;	// Supports a script interface
    flags.renderCommand		= 1;	// There is a UNIX command
    flags.renderMultiCPU	= 1;	// Renderer supports multi-CPU
    flags.renderMultiHost	= 0;	// Can't render over the network
    flags.renderMultiFrame	= 0;	// Can't render multiple frames at once
    flags.renderFields		= 0;	// No support for video fields

    // Now, here we have a suggested command string for Houdini
    flags.suggestedCommand	= "customrender -V2";
}

int
IFD_CustomRender::doSingleFrame(IFD_Scene &scene, IFD_RenderParms &rparms,
				float now)
{
    // Attempt to open a script or command.  The ostream returned will be
    //	either the result of a pipe open or a file open.
    if (!(os = open(rparms, now)))
	return 0;

    myObjectCount = scene.getNObjects(now);
    myLightCount  = scene.getNLights(now);
    myBinaryFlag  = rparms.myBinaryFlag;
    myStartTime = now;
    myEndTime = now;

    writeHeader(scene);
    outputLights(scene);
    outputObjects(scene);
    writeTrailer(scene);


    // Now, close the stream.  The base class method does the right thing.
    closeStream(os);

    return 1;
}

void
IFD_CustomRender::outputMatrix(const char *token, UT_DMatrix4 &mat)
{
    *os << token << " ";
    *os <<mat(0,0) <<" " << mat(0,1) <<" " << -mat(0,2) <<" " << mat(0,3)<<" ";
    *os <<mat(1,0) <<" " << mat(1,1) <<" " << -mat(1,2) <<" " << mat(1,3)<<" ";
    *os <<mat(2,0) <<" " << mat(2,1) <<" " << -mat(2,2) <<" " << mat(2,3)<<" ";
    *os <<mat(3,0) <<" " << mat(3,1) <<" " << -mat(3,2) <<" " << mat(3,3)<<"\n";
}

static inline float
computeZoom(IFD_CameraParms &parms)
{
    return parms.focal / parms.aperture;
}

void
IFD_CustomRender::writeHeader(IFD_Scene &scene)
{
    time_t		 now;
    UT_Vector3		 ambient;
    IFD_CameraParms	 parms;
    UT_String		 str;

    // Get information about the camera
    scene.getCameraParms(parms, myStartTime, renderName);

    // Get the camera transform so that we can flatten the transforms for
    //	all the other objects
    scene.getCameraTransform(myCameraTransform, myStartTime);
    myCameraTransform.invert();

    now = time(0);
    *os << "# Script created by Houdini: " << ctime(&now);
    *os << "# target is " << renderName << endl;

    // Now, find out the image that we're supposed to render to.
    scene.getOutputImage(str, myStartTime, renderName);
    *os << "picture " << (const char *)str << endl;

    *os << "samples " << parms.samplex << " " << parms.sampley << endl;
    *os << "resolution " << parms.xres << " " << parms.yres << endl;
    *os << "zoom " << computeZoom(parms) << endl;
    *os << "aspect_ratio " << parms.aspect << endl;

    if (parms.cropx[0] != 0 || parms.cropx[1] != 1 || parms.cropy[0] != 0 ||
		parms.cropy[1] != 1)
	*os << "crop "  << parms.cropx[0] << " " << parms.cropx[1] << " "
			<< parms.cropy[0] << " " << parms.cropy[1] << endl;

    *os << "zclip " << parms.mynear << " " << parms.myfar << endl;
}

void
IFD_CustomRender::outputLights(IFD_Scene &scene)
{
    int			i;
    IFD_LightParms	parms;
    IFD_CameraParms	cparms;
    UT_DMatrix4		xform;
    int			doreflect;

    *os << "# Light definitions" << endl;
    for (i = 0; i < myLightCount; i++)
    {
	scene.getLightParms(i, parms, myStartTime, renderName);
	scene.getLightParms(i, cparms, myStartTime, renderName);
	*os << "light " << parms.r << " " << parms.g << " " << parms.b;
	*os << " -z " << computeZoom(cparms);
	if (parms.isAtten) *os << " -a " << parms.atten;
	if (parms.isCone)
	{
	    *os << " -c " << parms.angle << " " << parms.delta <<
		   " -r " << parms.roll;
	}

	doreflect = 0;
	if (parms.reflectionMap.isstring())
	{
	    *os << " -s " << parms.reflectionMap;
	    doreflect = 1;
	}
	if (parms.diffuseMap.isstring())
	{
	    *os << " -d " << parms.reflectionMap;
	    doreflect = 1;
	}
	if (parms.slideMap.isstring())
	    *os << " -f " << parms.reflectionMap;
	if (parms.zdepthMap.isstring())
	    *os << " -Z " << parms.zdepthMap;

	if (doreflect)
	{
	    *os << " -R " << parms.reflectX << " " << parms.reflectY;
	}
	*os << endl;

	scene.getLightTransform(i, xform, myStartTime);
	xform *= myCameraTransform;
	outputMatrix("transform", xform);
    }
}

void
IFD_CustomRender::outputObjects(IFD_Scene &scene)
{
    int			 i, j;
    int			 gidx;
    UT_String		 name;
    UT_String		 gname;
    UT_String		 surface;
    UT_DMatrix4		 xform;
    IFD_GeometryParms	 parms;
    GU_Detail		*gdp;

    for (i = 0; i < myObjectCount; i++)
    {
	// Find out which geometry we're using - no, we don't
	//	support instancing
	gidx = scene.getObjectGeometry(i);

	// Output object token, name of object, detail, and material
	*os << "\n#  object ";
	scene.getObjectName(i, name, 1);
	if (name.isstring())	*os << name;
	else			*os << i;

	*os << "\ndetail ";
	scene.getGeometryParms(gidx, parms, myStartTime, renderName);
	if (parms.smoothShaded) *os << " -s";
	if (parms.backfaceRemove) *os << " -b";
	if (myBinaryFlag) *os << " stdin.bgeo";
	else		  *os << " stdin.geo";
	*os << endl;

	// Now, get that piece of geometry
	gdp = (GU_Detail *)scene.getGeometry(gidx, myStartTime);

	// If the renderer doesn't support Houdini .geometry files, you
	//	have to use a translator here.
	gdp->save(*os, myBinaryFlag);

	// Now, output the light mask
	*os << "light_mask ";
	for (j = 0; j < myLightCount; j++)
	{
	    if (scene.isObjectLit(i, j, myStartTime)) *os << " " << j;
	    else				 *os << " " << -j;
	}
	*os << endl;

	// Get the shader string for the object (see ../SHOP/SHOP_CustomClerk.C)
	surface = "";
	scene.getObjectShader(i, surface, myStartTime, renderName);
	if (surface.isstring())
	    *os << "surface " << (const char *)surface << endl;

	scene.getObjectTransform(i, xform, myStartTime);
	xform *= myCameraTransform;
	outputMatrix("transform", xform);
    }
}

void
IFD_CustomRender::writeTrailer(IFD_Scene &)
{
    *os << "render" << endl;
}


//
//  Now, we define the here is where we wire in the interface
//	to Houdini.

static IFD_Output *construct() { return (IFD_Output *)new IFD_CustomRender; }

void
IFDnewRenderer(void *)
{
    IFD_Entry	*custom;

    custom = new IFD_Entry(renderName, "Custom Render", ".pic", construct,
				0, 0, 0, 0);
    custom->insertIntoList();	// Make sure to add it to the list
}
