/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *	Side Effects Software Inc
 *	477 Richmond Street West
 *	Toronto, Ontario
 *	Canada   M5V 3E7
 *	416-504-9876
 *
 * NAME:	ROP_SimpleRib.h (  ROP Library, C++)
 *
 * COMMENTS:	This code is a ROP which uses the IFD_SimpleRib code (../IFD)
 *		to build a custom ROP for the render class.
 */

#include <stdlib.h>
#include <UT/UT_DSOVersion.h>
#include <CH/CH_LocalVariable.h>
#include <PRM/PRM_Include.h>
#include <PRM/PRM_DialogScript.h>
#include <OP/OP_Operator.h>
#include <OP/OP_OperatorTable.h>
#include <IFD/IFD_Entry.h>
#include <IFD/IFD_Table.h>
#include <IFD/IFD_Output.h>
#include "ROP_SimpleRib.h"

int			*ROP_SimpleRib::rmanIndirect = 0;
CH_LocalVariable	 ROP_SimpleRib::myVariableList[] = { {0, 0, 0} };

void
ROP_SimpleRib::setRendererParms(IFD_Output *)
{
    // Don't allow the use-display geometry
    useDisplayGeometry(0);
}

OP_Node *
ROP_SimpleRib::myConstructor(OP_Network *net, const char *name, OP_Operator *op)
{
    return new ROP_SimpleRib(net, name, op);
}

unsigned
ROP_SimpleRib::disableParms()
{
    int			 changed = 0;
    int			 backstate;

    backstate = 1;
    if (DORANGE()) backstate = 0;	// Don't allow background rendering
    if (DOSCRIPT()) backstate = 0;	// Don't allow background rendering

    changed += enableParm("command",	!DOSCRIPT());
    changed += enableParm("script",	 DOSCRIPT());
    changed += enableParm("binary",	 DOSCRIPT());
    changed += enableParm("resmenu",	 DORES());
    changed += enableParm("res",	 DORES());
    changed += enableParm("aspect",	 DORES());
    changed += enableParm("background",	backstate);

    // Allow the base class to disable parameters as well.
    return changed + ROP_Node::disableParms();
}

ROP_SimpleRib::ROP_SimpleRib(OP_Network *net, const char *name,
			 OP_Operator *entry) : ROP_IFD(net, name, entry)
{
    if (!rmanIndirect)
	rmanIndirect = allocIndirect(4);

    myRenderSuffix = ".tiff";
    myRenderDevice = "tiff";

    UT_String	rname("render");
    setString(rname, "renderer", 0, 0);
}

ROP_SimpleRib::~ROP_SimpleRib()
{
}

const char *
ROP_SimpleRib::getRenderName()
{
    return "RIB";
}

void
ROP_SimpleRib::getFrameDevice(UT_String &dev, float t, const char *)
{
    DEVICE(dev, t);
}

static IFD_Entry *
lookupIFDSimpleRib()
{
    IFD_Table	*table = IFDgetTable();
    IFD_Entry	*entry;

    for (entry = (IFD_Entry *)table->iterateInit(); entry;
	 entry = (IFD_Entry *)table->iterateNext(entry))
    {
	if (!strcmp(entry->name, "SimpleRib"))
	    return entry;
    }
    return 0;
}

IFD_Entry *
ROP_SimpleRib::getRenderType()
{
    return lookupIFDSimpleRib();
}

// The label for the commandName is quite important
static PRM_Name		commandName("command", "Render Command");
static PRM_Name		deviceName("device", "Frame Device");
static PRM_Default	DSrenderCommand	(PRM_DialogRenderers, "render");
static PRM_Default	defDevice(0, "framebuffer");

// Now define a menu for devices
static PRM_Name		deviceNames[] = {
			    PRM_Name("framebuffer",	"Frame Buffer"),
			    PRM_Name("mdisplay",	"Mantra Displayvice"),
			    PRM_Name("it",		"Persistent Display"),
			    PRM_Name("tiff",		"TIFF File"),
			    PRM_Name("zfile",		"Z Depth Image"),
			    PRM_Name("shadow",		"Shadow Format"),
			    PRM_Name("tga",		"Targa Format"),
			    PRM_Name("postscript",	"Postscript"),
			    PRM_Name("cpostscript",	"Color Postscript"),
			    PRM_Name("sgif",		"SGI Format"),
			    // PRM_Name("picio",	"PIXAR Pic Format"),
			    PRM_Name(),
};

static PRM_ChoiceList devMenu(PRM_CHOICELIST_REPLACE, deviceNames);

static PRM_Template	myrenParms[] = {
    // It is important that the command be the first parameter in this list.
    PRM_Template(PRM_COMMAND,	1, &commandName, &DSrenderCommand),
    PRM_Template(PRM_STRING,	1, &deviceName, &defDevice, &devMenu),

    PRM_Template(),
};

PRM_Template *
ROP_SimpleRib::getCommandTemplate()
{
    static int	firstTime = 1;
    if (firstTime)
    {
	// This code will prevent your IFD class from showing up
	// in the general scene renderer where it goes by default.
	firstTime = 0;
	registerSpecialRenderer("SimpleRib");
    }
    return myrenParms;
}

// This is the tricky part...

void
ROP_SimpleRib::installOperator(OP_OperatorTable *table)
{
    // Verify that the SimpleRib plug-in is in existence...
#if 0
    if (!lookupIFDSimpleRib())
	return;
#endif

    OP_VariablePair	*base_var = new OP_VariablePair(ROP_Node::myVariableList);
    OP_VariablePair	*myvar = new OP_VariablePair(ROP_IFD::myVariableList, base_var);

    PRM_Template	*myparms = ROP_SimpleRib::getCommandTemplate();

    OP_Operator	*op = new OP_Operator(
			    "simplerib", "Simple Rib",
			    ROP_SimpleRib::myConstructor,
			    ROP_Node::getROPifdTemplate(myparms),
			    0,
			    0,
			    myvar,
			    OP_FLAG_GENERATOR);
    table->addOperator(op);
}

void
newDriverOperator(OP_OperatorTable *table)
{
    ROP_SimpleRib::installOperator(table);
}
