/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *	Side Effects Software Inc.
 *      477 Richmond Street West
 *      Toronto, Ontario
 *      Canada   M5V 3E7
 *      416-504-9876
 *
 * NAME:	SOP Library (C++)
 *
 * COMMENTS:	Simple Particle demonstration code.  Includes sample
 *		collision detection...
 */

#include <UT/UT_DSOVersion.h>

#include <UT/UT_Math.h>
#include <UT/UT_Interrupt.h>
#include <UT/UT_Vector3.h>
#include <UT/UT_Vector4.h>

#include <GEO/GEO_PrimPart.h>

#include <GU/GU_Detail.h>
#include <GU/GU_RayIntersect.h>

#include <PRM/PRM_Include.h>

#include <OP/OP_Director.h>
#include <OP/OP_Operator.h>
#include <OP/OP_OperatorTable.h>

#include "SOP_SParticle.h"

void
newSopOperator(OP_OperatorTable *table)
{
     table->addOperator(new OP_Operator("sparticle",
					"Simple Particle",
					 SOP_SParticle::myConstructor,
					 SOP_SParticle::myTemplateList,
					 1,	// Min required sources
					 2,	// Maximum sources
					 0));
}

// The names here have to match the inline evaluation functions
static PRM_Name        names[] = {
    PRM_Name("reset",	"Reset Frame"),
    PRM_Name("birth",	"Birth Rate"),
    PRM_Name("force",	"Force"),
};

static PRM_Default	birthRate(10);

PRM_Template
SOP_SParticle::myTemplateList[] = {
    PRM_Template(PRM_INT,	1, &names[0], PRMoneDefaults),
    PRM_Template(PRM_INT_J,	1, &names[1], &birthRate),
    PRM_Template(PRM_XYZ_J,	3, &names[2]),
    PRM_Template(),
};

int *
SOP_SParticle::myOffsets = 0;

OP_Node *
SOP_SParticle::myConstructor(OP_Network *net, const char *name, OP_Operator *op)
{
    return new SOP_SParticle(net, name, op);
}

SOP_SParticle::SOP_SParticle(OP_Network *net, const char *name, OP_Operator *op)
	: SOP_Node(net, name, op)
{
    // Make sure that our offsets are allocated.  Here we allow up to 32
    // parameters, no harm in over allocating.  The definition for this
    // function is in OP/OP_Parameters.h
    if (!myOffsets)
	myOffsets = allocIndirect(32);

    // Now, flag that nothing has been built yet...
    myVelocity = -1;
}

SOP_SParticle::~SOP_SParticle() {}

unsigned
SOP_SParticle::disableParms()
{
    return 0;
}

void
SOP_SParticle::birthParticle()
{
    const GEO_Point	*src;
    GEO_Point		*ppt;
    GEO_ParticleVertex	*pvtx;
    float		*life;

    src = 0;
    pvtx = mySystem->giveBirth();
    if (mySource)
    {
	if (mySourceNum >= mySource->points().entries())
	    mySourceNum = 0;
	if (mySource->points().entries() > 0)	// No points in the input
	    src = mySource->points()(mySourceNum);
	mySourceNum++;		// Move on to the next source point...
    }
    if (src)
    {
	ppt = pvtx->getPt();
	ppt->getPos() = src->getPos();
	if (mySourceVel >= 0)
	{
	    memcpy(ppt->getAttribData(myVelocity),
		   src->getAttribData(mySourceVel), sizeof(UT_Vector3));
	}
	else memset(ppt->getAttribData(myVelocity), 0, sizeof(UT_Vector3));
    }
    else
    {
	ppt = pvtx->getPt();
	ppt->getPos().assign(drand48()-.5, drand48()-.5, drand48()-.5, 1);
	memset(ppt->getAttribData(myVelocity), 0, sizeof(UT_Vector3));
    }
    life = (float *)ppt->getAttribData(myLife);
    life[0] = 0;
    life[1] = 30+30*drand48();		// 30 frames + 30 random...
}

int
SOP_SParticle::moveParticle(GEO_ParticleVertex *pvtx, const UT_Vector3 &force)
{
    float	*life;
    float	 tinc;
    UT_Vector3	*vel, dir;

    life = (float *)pvtx->getPt()->getAttribData(myLife);
    life[0] += 1;
    if (life[0] >= life[1])
	return 0;		// The particle should die!

    tinc = 1./30.;		// Hardwire 1/30 of a second time inc...

    // Adjust the velocity (based on the force) - of course, the multiplies
    // can be pulled out of the loop...
    vel = (UT_Vector3 *)pvtx->getPt()->getAttribData(myVelocity);
    vel->x() += tinc*force.x();
    vel->y() += tinc*force.y();
    vel->z() += tinc*force.z();

    // Now adjust the point positions

    if (myCollision)
    {
	dir = *vel;
	dir *= tinc;

	// here, we only allow hits within the length of the velocity vector
	GU_RayInfo	info(dir.normalize());
	UT_Vector3	start;

	start = pvtx->getPos();
	if (myCollision->sendRay(start, dir, info) > 0)
	    return 0;	// We hit something, so kill the particle
    }

    pvtx->getPos().x() += tinc*vel->x();
    pvtx->getPos().y() += tinc*vel->y();
    pvtx->getPos().z() += tinc*vel->z();

    return 1;
}

void
SOP_SParticle::timeStep(float now)
{
    int			 i, nbirth;
    GEO_ParticleVertex	*pvtx, *next = 0;
    UT_Vector3		 force;

    force.assign(FX(now), FY(now), FZ(now));
    nbirth = BIRTH(now);

    if (error() >= UT_ERROR_ABORT)
	return;

    for (i = nbirth; i >= 0; i--)
	birthParticle();

    for (pvtx = mySystem->iterateInit(); pvtx; pvtx = next)
    {
	next = mySystem->iterateFastNext(pvtx);
	if (!moveParticle(pvtx, force))
	    mySystem->deadParticle(pvtx);
    }
}

void
SOP_SParticle::initSystem()
{
    if (!gdp) gdp = new GU_Detail;

    // Check to see if we really need to reset everything
    if (gdp->points().entries() > 0 || myVelocity < 0)
    {
	mySourceNum = 0;
	gdp->clearAndDestroy();
	mySystem = (GEO_PrimParticle *)gdp->appendPrimitive(GEOPRIMPART);
	mySystem->clearAndDestroy();

	// A vector attribute will be transformed correctly by following
	//	SOPs.  Use float types for stuff like color...
	myVelocity = gdp->addPointAttrib("v", sizeof(UT_Vector3),
					 GB_ATTRIB_VECTOR, 0);
	myLife = gdp->addPointAttrib("life", sizeof(float)*2,
					 GB_ATTRIB_FLOAT, 0);
    }
}

OP_ERROR
SOP_SParticle::cookMySop(OP_Context &context)
{
    float		 reset, currframe;
    CH_Manager		*chman;		// Channel manager has time info for us
    const GU_Detail	*collision;

    // Before we do anything, we must lock our inputs.  Before returning,
    //	we have to make sure that the inputs get unlocked.
    if (lockInputs(context) >= UT_ERROR_ABORT)
	return error();

    // Now, indicate that we are time dependent (i.e. have to cook every
    //	time the current frame changes).
    OP_Node::flags().timeDep = 1;

    chman = OPgetDirector()->getChannelManager();

    // This is the frame that we're cooking at...
    currframe = chman->getSample(context.myTime);
    reset = RESET();		// Find our reset frame...

    if (currframe <= reset)
    {
	myLastCookTime = reset;
	initSystem();
    }
    else
    {
	// Set up the collision detection object
	collision = inputGeo(1, context);
	if (collision)
	{
	    myCollision = new GU_RayIntersect;
	    myCollision->init(collision);
	}
	else myCollision = 0;

	// Set up our source information...
	mySource = inputGeo(0, context);
	if (mySource)
	{
	    mySourceVel = mySource->findPointAttrib("v", sizeof(UT_Vector3),
				    GB_ATTRIB_VECTOR);

	    // If there's no velocity, pick up the velocity from the normal
	    if (mySourceVel < 0)
		mySourceVel = mySource->findPointAttrib("N", sizeof(UT_Vector3),
				    GB_ATTRIB_VECTOR);
	}

	// This is where we notify our handles (if any) if the inputs have
	// changed.  Since there is no input group, "0" is passed as the
	// fourth argument.
	checkInputChanged(0, -1, myDetailGroupPair, (GU_Detail *)mySource, 0);

	// Now cook the geometry up to our current time
	// Here, we could actually re-cook the source input to get a moving
	//	source...  But this is just an example ;-)

	currframe += 0.05;	// Add a bit to avoid floating point error
	while (myLastCookTime < currframe)
	{
	    // Here we have to convert our frame number to the actual time.
	    timeStep(chman->getTime(myLastCookTime));
	    myLastCookTime += 1;
	}

	if (myCollision) delete myCollision;

	// Highlight all the birthed particles.  This call also clears the
	// selection at the beginning.  The argument "GU_SPoint" indicates
	// that we want a point selection.
	select(GU_SPoint);
    }

    unlockInputs();
    return error();
}

const char *
SOP_SParticle::inputLabel(unsigned inum) const
{
    switch (inum)
    {
	case 0: return "Particle Source Geometry";
	case 1:	return "Collision Object";
    }
    return "Unknown source";
}
