/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *      Side Effects
 *      477 Richmond Street West
 *      Toronto, Ontario
 *      Canada   M5V 3E7
 *      416-504-9876
 *
 * NAME:	SOP Library (C++)
 *
 * COMMENTS:	The Star SOP
 */

#include <limits.h>
#include <UT/UT_DSOVersion.h>
#include <UT/UT_Math.h>
#include <UT/UT_Interrupt.h>
#include <GU/GU_Detail.h>
#include <GU/GU_PrimPoly.h>
#include <CH/CH_LocalVariable.h>
#include <PRM/PRM_Include.h>
#include <OP/OP_Operator.h>
#include <OP/OP_OperatorTable.h>
#include "SOP_Star.h"

//
// Here, we sub-class off of OP_Operator so that we can embed our help in the C
// file.
//
class OP_StarOperator : public OP_Operator {
public:
	     OP_StarOperator();
    virtual ~OP_StarOperator();

    virtual void	getOperatorHelp(const char *pathname, UT_String &str,
					int level, bool *is_html_ptr,
					bool try_other_levels) const;
};

OP_StarOperator::OP_StarOperator()
    : OP_Operator("star",
		  "Star",
		  SOP_Star::myConstructor,
		  SOP_Star::myTemplateList,
		  0,
		  0,
		  SOP_Star::myVariables,
		  OP_FLAG_GENERATOR)
{
}

OP_StarOperator::~OP_StarOperator()
{
}

void
OP_StarOperator::getOperatorHelp(const char *, UT_String &help, int /* level */,
				 bool *is_html_ptr,
				 bool /* try_other_levels */) const
{
    help  = "Star SOP builds stars of N points\n\n";

    help += "   Divisions:   Number of points on the star\n";
    help += "   Radius:      Radius of the star\n";
    help += "   NegRadius:   Allow the radius to be negative\n";
    help += "                This may create bowtie polygons\n";
    help += "   Center:      The center point of the star\n";
    help += "   Orientaion:  The orientation of the star\n";
    help += "\n";
    help += "Local Variables:\n";
    help += "  $PT  - current point number\n";
    help += "  $NPT - number of points in the star\n";

    // If they're asking if this is html help, it's not.
    if (is_html_ptr)
	is_html_ptr = false;
}

void
newSopOperator(OP_OperatorTable *table)
{
#if 0
    // If we want to use the "standard" operator class, we can do this here.
    // However, the OP_StarOperator has the help built in (and doesn't require
    // an external file).
    table->addOperator(
	    new OP_Operator("star",			// Internal name
			    "Star",			// UI name
			     SOP_Star::myConstructor,	// How to build the SOP
			     SOP_Star::myTemplateList,	// My parameters
			     0,				// Min # of sources
			     0,				// Max # of sources
			     SOP_Star::myVariables,	// Local variables
			     OP_FLAG_GENERATOR)		// Flag it as generator
	    );
#else
     table->addOperator(new OP_StarOperator());
#endif
}

static PRM_Name		negativeName("nradius", "Negative Radius");
//				     ^^^^^^^^    ^^^^^^^^^^^^^^^
//				     internal    descriptive version

static PRM_Default	fiveDefault(5);		// Default to 5 divisions
static PRM_Default	radiiDefaults[] = {
			   PRM_Default(1),		// Outside radius
			   PRM_Default(0.3)		// Inside radius
			};

PRM_Template
SOP_Star::myTemplateList[] = {
    PRM_Template(PRM_INT_J,	1, &PRMdivName, &fiveDefault, 0,
					&PRMdivision2Range),
    PRM_Template(PRM_XYZ_J,	2, &PRMradiusName, radiiDefaults),
    PRM_Template(PRM_TOGGLE,	1, &negativeName),
    PRM_Template(PRM_XYZ_J,     3, &PRMcenterName),
    PRM_Template(PRM_ORD,	1, &PRMorientName, 0, &PRMplaneMenu),
    PRM_Template()
};


// Here's how we define local variables for the SOP.
enum {
	VAR_PT,		// Point number of the star
	VAR_NPT		// Number of points in the star
};

CH_LocalVariable
SOP_Star::myVariables[] = {
    { "PT",	VAR_PT, 0 },		// The table provides a mapping
    { "NPT",	VAR_NPT, 0 },		// from text string to integer token
    { 0, 0, 0 },
};

float
SOP_Star::getVariableValue(int index, int)
{
    // Note that "gdp" may be null here, so we do the safe thing
    // and cache values we are interested in.
    if (myCurrPoint < 0) return 0;
    switch (index)
    {
	case VAR_PT:	return myCurrPoint;
	case VAR_NPT:	return myTotalPoints;
    }
    return 0;		// Whoops, not one of our variables
}

OP_Node *
SOP_Star::myConstructor(OP_Network *net, const char *name, OP_Operator *op)
{
    return new SOP_Star(net, name, op);
}

SOP_Star::SOP_Star(OP_Network *net, const char *name, OP_Operator *op)
	: SOP_Node(net, name, op)
{
    myCurrPoint = -1;		// To prevent garbage values from being returned
}

SOP_Star::~SOP_Star() {}

unsigned
SOP_Star::disableParms()
{
    return 0;
}

OP_ERROR
SOP_Star::cookMySop(OP_Context &context)
{
    float		 now;
    float		 rad, tx, ty, tz;
    int			 divisions, plane, negradius;
    int			 xcoord, ycoord, zcoord;
    float		 tmp, tinc;
    GEO_Point		*ppt;
    GU_PrimPoly		*poly;
    int			 i;
    UT_Interrupt	*boss;

    now = context.myTime;

    // Since we don't have inputs, we don't need to lock them.

    divisions  = DIVISIONS(now)*2;	// We need twice our divisions of points
    myTotalPoints = divisions;		// Set the NPT local variable value
    myCurrPoint   = 0;			// Initialize the PT local variable

    plane      = ORIENT();
    negradius  = NEGRADIUS();
    tx	       = CENTERX(now);
    ty	       = CENTERY(now);
    tz	       = CENTERZ(now);

    switch (plane)
    {
	case 0:		// XY Plane
		xcoord = 0;
		ycoord = 1;
		zcoord = 2;
		break;
	case 1:		// YZ Plane
		xcoord = 1;
		ycoord = 2;
		zcoord = 0;
		break;
	case 2:		// XZ Plane
		xcoord = 0;
		ycoord = 2;
		zcoord = 1;
		break;
    }

    // Check to see that there hasn't been a critical error in cooking the SOP.
    if (error() < UT_ERROR_ABORT)
    {
	boss = UTgetInterrupt();
	if (divisions < 4)
	{
	    // With the range restriction we have on the divisions, this
	    //	is actually impossible, but it shows how to add an error
	    //	message or warning to the SOP.
	    addWarning(SOP_MESSAGE, "Invalid divisions");
	    divisions = 4;
	}
	gdp->clearAndDestroy();

	// Start the interrupt server
	boss->opStart("Building Star");
	// Build a polygon
	poly = GU_PrimPoly::build(gdp, divisions, GU_POLY_CLOSED);
	tinc = M_PI*2 / (float)divisions;

	// Now, set all the points of the polygon
	for (i = 0; i < divisions; i++)
	{
	    // Check to see if the user has interrupted us...
	    if (boss->opInterrupt())
		break;

	    myCurrPoint = i;

	    // Since we expect the local variables to be used in specifying
	    // the radii, we have to evaluate the channels INSIDE the loop
	    // through the points...

	    tmp = (float)i * tinc;
	    rad = (i & 1) ? XRADIUS(now) : YRADIUS(now);
	    if (!negradius && rad < 0)
		rad = 0;

	    ppt = poly->getVertex(i).getPt();
	    ppt->getPos()(xcoord) = cos(tmp) * rad + tx;
	    ppt->getPos()(ycoord) = sin(tmp) * rad + ty;
	    ppt->getPos()(zcoord) = 0 + tz;
	}

	// Highlight the star which we have just generated.  This routine call
	// clears any currently highlighted geometry, and then it highlights 
	// every primitive for this SOP. 
	select(GU_SPrimitive);

	// Tell the interrupt server that we've completed
	boss->opEnd();
    }

    myCurrPoint = -1;
    return error();
}
