/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *	Mark Elendt
 *	Side Effects Software Inc
 *	477 Richmond Street West
 *	Toronto, Ontario
 *	Canada   M5V 3E7
 *	416-504-9876
 *
 * NAME:	traverse.C ( C++)
 *
 * COMMENTS:	This sample program traverses the whole heirarchy printing out
 *		information about the nodes in the file.
 */


#include <fstream.h>
#include <iostream.h>
#include <UT/UT_NTStreamUtil.h>
#include <CMD/CMD_Args.h>
#include <MOT/MOT_Director.h>

static int	verbose = 0;

static void
printIndent(int indent)
{
    while (indent > 0)
    {
	cout << " ";
	indent--;
    }
}

static void
printNode(OP_Node *node, OP_Node *input, int indent)
{
    int		 i, noutputs;
    OP_Node	*out;

    printIndent(indent);
    cout << node->getName();
    if (input)
	cout << " is wired to " << input->getName();
    cout << endl;

    // Now, we print out the nodes which have this node as an input
    noutputs = node->nOutputs();
    for (i = 0; i < noutputs; i++)
    {
	out = node->getOutput(i);
	printNode(out, node, indent+1);
    }
}

static void
printNetwork(OP_Network *net, int indent = 0)
{
    int		 i, nkids;
    OP_Node	*node;
    OP_Network	*kidnet;

    // Here, we find all the nodes which don't have inputs.  These are the
    // "root" nodes.  We follow the outputs down for each node printing out the
    // information.
    nkids = net->getNchildren();
    for (i = 0; i < nkids; i++)
    {
	node = net->getChild(i);

	// If we have no inputs, then print us out, otherwise, the node
	// printing will print us out...
	if (node->nInputs() == 0)
	    printNode(node, 0, indent);
    }

    // Now, if any of the nodes are networks themselves, let's print out them
    // and their contents.

    for (i = 0; i < nkids; i++)
    {
	node = net->getChild(i);
	if (node->isNetwork())
	{
	    // Here, this is a safe cast.
	    kidnet = (OP_Network *)node;
	    if (kidnet->getNchildren())
	    {
		printIndent(indent+1);
		cout << kidnet->getName() << " contains:" << endl;
		printNetwork(kidnet, indent+2);
	    }
	}
    }
}

int
main(int argc, char *argv[])
{
    CMD_Args		 args;
    MOT_Director	*boss;
    int			 i;

    boss = new MOT_Director("traverse");
    OPsetDirector(boss);

    args.initialize(argc, argv);
    args.stripOptions("v");

    if (args.found('v'))
	verbose = 1;

    // Load the arguments
    for (i = 1; i < args.argc(); i++)
    {
	ifstream	is(args(i), ios::in | NTBinaryMode | UT_NoCreate);
	if (is)
	{
	    // Merge in all the arguments
	    if (verbose)
		cout << "Loading: " << args(i) << endl;
	    if (!boss->loadNetwork(is, 0, 1))
		cerr << "Error loading network for " << args(i) << endl;
	}
    }

    if (verbose)
	cout << "Traversing the HIP file(s)" << endl;

    printNetwork(boss);

    return 0;
}
